//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "common.h"
//---------------------------------------------------------------------------
int parseNumber(char *str);
int readFile(char *fileName, BYTE *buff, int minLen, int maxLen);
int openFile(char *fileName, int minLen, int maxLen, FILE **fr);
int writeFile(char *fileName, BYTE *buff, int len, BOOL append);
int printError(int err, void *par);
void version(void);
void usage(void);

//---------------------------------------------------------------------------
int main(int argc, char* argv[])
{
  char *fnIn = NULL, *fnOut = NULL, fn[256], n[9], *pc, fileType = 0;
  BOOL append = FALSE, hless = FALSE, basic = FALSE;
  int beginAddr = -1, fileNum = -1;
  BYTE buff1[65536], buff2[65536], *pb;
  int i, err, lenIn, lenOut;
  FILE *f = NULL;

  // spracovanie a kontrola vstupnych parametrov
  *n = 0;
  err = 0;
  do {
    for (i = 1; i < argc; i++) {
      if (strcmp(argv[i], "-h") == 0) {
        err = ERR_PRINT_USAGE;
        break;
      }
      else if (strcmp(argv[i], "-v") == 0) {
        err = ERR_PRINT_VERSION;
        break;
      }
      else if (strcmp(argv[i], "-append") == 0) {
        append = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-hless") == 0) {
        hless = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-b") == 0) {
        basic = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-a") == 0) {
        if (++i == argc)
          beginAddr = -1;
        else
          beginAddr = parseNumber(argv[i]);
        if (beginAddr < 0 || beginAddr > 65535) {
          err = printError(ERR_BAD_ADDRESS, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-n") == 0) {
        if (++i == argc)
          fileNum = -1;
        else
          fileNum = parseNumber(argv[i]);
        if (fileNum < 0 || fileNum > 99) {
          err = printError(ERR_BAD_FILE_NUMBER, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-t") == 0) {
        if (++i == argc || strlen(argv[i]) > 1)
          fileType = 0;
        else
          fileType = argv[i][0];
        if (fileType < '!' || fileType > '~') {
          err = printError(ERR_BAD_FILE_TYPE, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-m") == 0) {
        if (++i == argc || strlen(argv[i]) > 8)
          n[0] = -1;
        else {
          strcpy(n, argv[i]);
          while (strlen(n) < 8)
            strcat(n, " ");
        }
        if (n[0] == -1) {
          err = printError(ERR_BAD_FILE_NAME, argv[i - 1]);
          break;
        }
      }
      else if (strcmp(argv[i], "-o") == 0) {
        if (++i == argc || fnOut != NULL) {
          err = printError(ERR_OUTPUT_FILE_MORE, NULL);
          break;
        }
        fnOut = argv[i];
        continue;
      }
      else {
        if (fnIn != NULL) {
          err = printError(ERR_INPUT_FILE_MORE, NULL);
          break;
        }
        fnIn = argv[i];
        continue;
      }
    }
    if (err != 0)
      break;

    // kontrola povolenych kombinacii volieb
    if ((hless && (basic || beginAddr >= 0 || fileType != 0 || fileNum >= 0))
        || (basic && (beginAddr >= 0 || fileType != 0))) {
      err = printError(ERR_BAD_OPTIONS, NULL);
      break;
    }

    // kontrola vstupneho suboru
    if (fnIn == NULL) {
      err = printError(ERR_NO_INPUT_FILE, NULL);
      break;
    }

    // nazov vystupneho suboru
    if (fnOut == NULL) {
      strcpy(fn, fnIn);
      strcat(fn, ".ptp");
      fnOut = fn;
    }

    // default hodnoty
    if (beginAddr < 0)
      beginAddr = (basic) ? 0x2401 : 0;
    if (fileNum < 0)
      fileNum = 0;
    if (fileType == 0)
      fileType = (char) ((basic) ? '>' : '?');

    // meno suboru v PTP
    if (n[0] == 0) {
      pc = strrchr(fnOut, '/');
      if (pc == NULL)
        pc = strrchr(fnOut, '\\');
      if (pc == NULL)
        pc = fnOut;
      else
        pc++;
      if (strlen(pc) > 0) {
        strncpy(n, pc, 8);
        n[8] = 0;
      }
      else
        strcpy(n, pc);
      pc = strrchr(n, '.');
      if (pc != NULL)
        *pc = 0;
      while (strlen(n) < 8)
        strcat(n, " ");
    }

    // spracuj vstupny subor
    pb = buff1;
    if (basic) {
      err = openFile(fnIn, 5, 65535, &f);
      if (err < 0)
        break;
      lenIn = parseBasicProgram(buff2, f);
      constructHeaderBlock(&pb, buff2, lenIn, fileType, fileNum, n, beginAddr);
    }
    else if (hless) {
      err = readFile(fnIn, buff2, 1, 65533);
      if (err < 0)
        break;
      lenIn = err;
      constructHeaderlessBlock(&pb, buff2, lenIn);
    }
    else {
      err = readFile(fnIn, buff2, 1, 65533);
      if (err < 0)
        break;
      lenIn = err;
      constructHeaderBlock(&pb, buff2, lenIn, fileType, fileNum, n, beginAddr);
    }
    lenOut = pb - buff1;

    // zapis vystupneho suboru
    printf("Writing to the output file '%s'\n", fnOut);
    err = writeFile(fnOut, buff1, lenOut, append);

  } while (FALSE);

  if (f != NULL)
    fclose(f);

  if (err == ERR_PRINT_USAGE) {
    usage();
    err = 0;
  }
  if (err == ERR_PRINT_VERSION) {
    version();
    err = 0;
  }

  return -err;
}
//---------------------------------------------------------------------------
int parseNumber(char *str)
{
  char *p;
  int val = strtol(str, &p, 0);
  if (val < 0 || (val == 0 && *p != 0))
    val = -1;
  return val;
}
//---------------------------------------------------------------------------
int readFile(char *fileName, BYTE *buff, int minLen, int maxLen)
{
  int err = 0, len;
  FILE *f;

  do {
    f = fopen(fileName, "rb");
    if (f == NULL) {
      err = printError(ERR_OPEN_SRC_FILE, fileName);
      break;
    }
    if (fseek(f, 0L, SEEK_END) != 0) {
      err = printError(ERR_ACCESS_SRC_FILE, fileName);
      break;
    }
    len = ftell(f);
    if (len < 0) {
      err = printError(ERR_LENGTH_SRC_FILE, fileName);
      break;
    }
    if (len < minLen || len > maxLen) {
      err = printError(ERR_BAD_LEN_SRC_FILE, fileName);
      break;
    }
    if (fseek(f, 0L, SEEK_SET) != 0) {
      err = printError(ERR_ACCESS_SRC_FILE, fileName);
      break;
    }
    if (fread(buff, 1, len, f) != (unsigned int)len) {
      err = printError(ERR_READ_SRC_FILE, fileName);
      break;
    }
  } while (FALSE);

  fclose(f);

  if (err < 0)
    return err;

  return len;
}
//---------------------------------------------------------------------------
int openFile(char *fileName, int minLen, int maxLen, FILE **fr)
{
  int err = 0, len;
  FILE *f;

  do {
    f = fopen(fileName, "rb");
    if (f == NULL) {
      err = printError(ERR_OPEN_SRC_FILE, fileName);
      break;
    }
    if (fseek(f, 0L, SEEK_END) != 0) {
      err = printError(ERR_ACCESS_SRC_FILE, fileName);
      break;
    }
    len = ftell(f);
    if (len < 0) {
      err = printError(ERR_LENGTH_SRC_FILE, fileName);
      break;
    }
    if (len < minLen || len > maxLen) {
      err = printError(ERR_BAD_LEN_SRC_FILE, fileName);
      break;
    }
    if (fseek(f, 0L, SEEK_SET) != 0) {
      err = printError(ERR_ACCESS_SRC_FILE, fileName);
      break;
    }
  } while (FALSE);

  if (err < 0) {
    fclose(f);
    return err;
  }

  *fr = f;
  return len;
}
//---------------------------------------------------------------------------
int writeFile(char *fileName, BYTE *buff, int len, BOOL append)
{
  int err = 0;
  FILE *f;

  do {
    if (append)
      f = fopen(fileName, "ab");
    else
      f = fopen(fileName, "wb");
    if (f == NULL) {
      err = printError(ERR_OPEN_DEST_FILE, fileName);
      break;
    }
    if (!append && fseek(f, 0L, SEEK_SET) != 0) {
      err = printError(ERR_ACCESS_DEST_FILE, fileName);
      break;
    }
    if (fwrite(buff, 1, len, f) != (unsigned int)len) {
      err = printError(ERR_WRITE_DEST_FILE, fileName);
      break;
    }
  } while (FALSE);

  fclose(f);

  return err;
}
//---------------------------------------------------------------------------
int printError(int err, void *par)
{
  switch (err) {
    case ERR_BAD_OPTIONS :
      fprintf(stderr, "Bad combination of options!");
      break;

    case ERR_OPEN_SRC_FILE :
      fprintf(stderr, "Can't open input file '%s'!", (char *)par);
      break;

    case ERR_ACCESS_SRC_FILE :
      fprintf(stderr, "Can't access input file '%s'!", (char *)par);
      break;

    case ERR_LENGTH_SRC_FILE :
      fprintf(stderr, "Can't retrieve length of the input file '%s'!", (char *)par);
      break;

    case ERR_BAD_LEN_SRC_FILE :
      fprintf(stderr, "Length of the input file '%s' must be 16384!", (char *)par);
      break;

    case ERR_READ_SRC_FILE :
      fprintf(stderr, "Can't read from the input file '%s'!", (char *)par);
      break;

    case ERR_OPEN_DEST_FILE :
      fprintf(stderr, "Can't open output file '%s'!", (char *)par);
      break;

    case ERR_ACCESS_DEST_FILE :
      fprintf(stderr, "Can't access output file '%s'!", (char *)par);
      break;

    case ERR_WRITE_DEST_FILE :
      fprintf(stderr, "Can't write to the output file '%s'!", (char *)par);
      break;

    case ERR_BAD_ADDRESS :
      fprintf(stderr, "After '%s' must follow address in range 0 to 65535!", (char *)par);
      break;

    case ERR_BAD_FILE_NUMBER :
      fprintf(stderr, "After '%s' must follow file number in range 0 to 99!", (char *)par);
      break;

    case ERR_BAD_FILE_TYPE :
      fprintf(stderr, "After '%s' must follow file type as single ASCII character!", (char *)par);
      break;

    case ERR_BAD_FILE_NAME :
      fprintf(stderr, "After '%s' must follow file name (max. 8 characters)!", (char *)par);
      break;

    case ERR_NO_INPUT_FILE :
      fprintf(stderr, "Input file not specified!\nTry '%s -h' for help.", PROG_EXE_NAME);
      break;

    case ERR_INPUT_FILE_MORE :
      fprintf(stderr, "Only one input file can be specified!");
      break;

    case ERR_OUTPUT_FILE_MORE :
      fprintf(stderr, "Only one output file can be specified!");
      break;
  }

  fprintf(stderr, "\n\n");

  return err;
}
//---------------------------------------------------------------------------
void version(void)
{
  printf("%s v%d.%d - %s\n",
         PROG_NAME, PROG_VERSION / 16, PROG_VERSION % 16, PROG_AUTHOR);
}
//---------------------------------------------------------------------------
void usage(void)
{
  version();

  printf("Usage: %s [options] infile.ext\n", PROG_EXE_NAME);
  printf("\nCommon options:\n");
  printf("  -o outfile.ext  output file name\n");
  printf("  -append\t  append at the end of the output file\n");
  printf("  -a address\t  begin address [0]\n");
  printf("  -n number\t  file number [0]\n");
  printf("  -t char\t  file type [?]\n");
  printf("  -m name\t  file name max. 8 chars [infile name]\n");
  printf("  -hless\t  store as headless block\n");
  printf("  -b\t\t  import BASIC program from TXT file\n");
  printf("  -h\t\t  usage information\n");
  printf("  -v\t\t  version info\n");
  printf("\nAll numeric parameters must be nonnegative integers and can be in three forms:\n");
  printf("  decimal (e.g. 12345)\n");
  printf("  octal (e.g. 01234)\n");
  printf("  hexadecimal (e.g. 0x1234)\n\n");

  return;
}
//---------------------------------------------------------------------------

